""" optionally builds inside polygon algorithm C extension

Resources:
https://github.com/FirefoxMetzger/mini-extension
https://stackoverflow.com/questions/60073711/how-to-build-c-extensions-via-poetry
https://github.com/libmbd/libmbd/blob/master/build.py
"""
import pathlib
import re
import warnings

import cffi
import cffi.setuptools_ext
import setuptools

EXTENSION_NAME = "inside_polygon_ext"
H_FILE_NAME = "inside_polygon_int.h"
C_FILE_NAME = "inside_polygon_int.c"
# stored in "timezonefinder" package folder for cleaner footprint
EXTENSION_PATH = pathlib.Path().resolve() / "timezonefinder" / "inside_poly_extension"
h_file_path = EXTENSION_PATH / H_FILE_NAME
c_file_path = EXTENSION_PATH / C_FILE_NAME

try:
    ffibuilder = cffi.FFI()
except Exception as exc:
    warnings.warn(f"C lang extension cannot be build, since cffi failed with this error: {exc}")
    # Clang extension should be fully optional
    ffibuilder = None

if ffibuilder is not None:
    ffibuilder.set_source(
        EXTENSION_NAME,  # name of the output C extension
        f'#include "{h_file_path}"',
        sources=[str(c_file_path)],
    )

    with open(h_file_path) as h_file:
        # cffi does not like our preprocessor directives, so we remove them
        lns = h_file.read().splitlines()
        flt = filter(lambda ln: not re.match(r" *#", ln), lns)

    ffibuilder.cdef("\n".join(flt))

    # with open(c_file_path) as c_file:
    #     # cffi does not like our preprocessor directives, so we remove them
    #     c_file_content = c_file.read()


def build_c_extension():
    if ffibuilder is None:
        warnings.warn(f"missing ffibuilder. skipping build process")
        return

    # not required
    # ffibuilder.compile(verbose=True)

    # Note: built into "timezonefinder" package folder
    distribution = setuptools.Distribution({"package_dir": {"": "timezonefinder"}})
    cffi.setuptools_ext.cffi_modules(distribution, "cffi_modules", ["build.py:ffibuilder"])
    cmd = distribution.cmdclass["build_ext"](distribution)
    cmd.inplace = 1
    cmd.ensure_finalized()
    try:
        cmd.run()
    except Exception as exc:
        # distutils.errors.CompileError:
        # a build failure in the extension (e.g. C compile is not installed) must not abort the build process,
        # but instead simply not install the failing extension.
        warnings.warn(f"C lang extension cannot be build, since cmd.run() failed with this error: {exc}")


if __name__ == "__main__":
    build_c_extension()



"""modify PYTHONPATH to make parent package discoverable."""

import sys
from os.path import abspath, join, pardir

package_path = abspath(join(__file__, pardir, pardir))
sys.path.insert(0, package_path)



from timezonefinder.timezonefinder import TimezoneFinder, TimezoneFinderL

# https://docs.python.org/3/tutorial/modules.html#importing-from-a-package
# determines which objects will be imported with "import *"
__all__ = ("TimezoneFinder", "TimezoneFinderL")



import struct
from pathlib import Path
from typing import Dict, List

import numpy as np
from h3.api import numpy_int as h3

from timezonefinder.configs import (
    DTYPE_FORMAT_B,
    DTYPE_FORMAT_H,
    DTYPE_FORMAT_H_NUMPY,
    DTYPE_FORMAT_Q,
    NR_BYTES_B,
    NR_BYTES_I,
    NR_BYTES_Q,
    THRES_DTYPE_B,
    ShortcutMapping,
)


def export_shortcuts_binary(global_mapping: Dict[int, List[int]], path2shortcuts: Path) -> int:
    """
    binary format:
        for every shortcut entry:
            - the hex id (uint64)
            - the amount of contained polygons n (uint8)
            - n polygon ids (uint16)

    """
    shortcut_space = 0
    with open(path2shortcuts, "wb") as fp:
        for hex_id, poly_ids in global_mapping.items():
            fp.write(struct.pack(DTYPE_FORMAT_Q, hex_id))
            nr_polys = len(poly_ids)
            if nr_polys > THRES_DTYPE_B:
                raise ValueError("value overflow: more polys than data type supports")
            fp.write(struct.pack(DTYPE_FORMAT_B, nr_polys))
            for poly_id in poly_ids:
                fp.write(struct.pack(DTYPE_FORMAT_H, poly_id))

            shortcut_space += NR_BYTES_Q + NR_BYTES_B + (len(poly_ids) * NR_BYTES_I)

    return shortcut_space


def read_shortcuts_binary(path2shortcuts: Path) -> ShortcutMapping:
    mapping: ShortcutMapping = {}
    with open(path2shortcuts, "rb") as fp:
        while 1:
            try:
                hex_id: int = struct.unpack(DTYPE_FORMAT_Q, fp.read(NR_BYTES_Q))[0]
            except struct.error:
                # EOF: buffer not long enough to unpack
                break
            nr_polys: int = struct.unpack(DTYPE_FORMAT_B, fp.read(NR_BYTES_B))[0]
            poly_ids: np.ndarray = np.fromfile(fp, dtype=DTYPE_FORMAT_H_NUMPY, count=nr_polys)
            mapping[hex_id] = poly_ids

    return mapping


def lies_in_h3_cell(h: int, lng: float, lat: float) -> bool:
    res = h3.h3_get_resolution(h)
    return h3.geo_to_h3(lat, lng, res) == h



""" 'transparent' numba functionality replacements

njit decorator
data types

dtype_2int_tuple = typeof((1, 1))
@njit(b1(i4, i4, i4[:, :]), cache=True)
@njit(dtype_2int_tuple(f8, f8), cache=True)
"""


def njit(*args, **kwargs):
    def wrapper(f):
        return f

    return wrapper


class SubscriptAndCallable:
    def __init__(self, *args, **kwargs):
        pass

    def __class_getitem__(cls, item):
        return None


# DTYPES
# @njit(b1(i4, i4, i4[:, :]), cache=True)


class b1(SubscriptAndCallable):
    pass


class f8(SubscriptAndCallable):
    pass


class i2(SubscriptAndCallable):
    pass


class i4(SubscriptAndCallable):
    pass


class u2(SubscriptAndCallable):
    pass
